@@ -5,8 +5,8 @@
  *
  * Joe Conway <mail@joeconway.com>
  *
- * $PostgreSQL: pgsql/contrib/fuzzystrmatch/fuzzystrmatch.c,v 1.22 2006/07/10 18:40:16 momjian Exp $
- * Copyright (c) 2001-2006, PostgreSQL Global Development Group
+ * $PostgreSQL: pgsql/contrib/fuzzystrmatch/fuzzystrmatch.c,v 1.30 2009/06/11 14:48:51 momjian Exp $
+ * Copyright (c) 2001-2009, PostgreSQL Global Development Group
  * ALL RIGHTS RESERVED;
  *
  * levenshtein()
@@ -15,6 +15,8 @@
  * found at http://www.merriampark.com/ld.htm
  * Also looked at levenshtein.c in the PHP 4.0.6 distribution for
  * inspiration.
+ * Configurable penalty costs extension is introduced by Volkan
+ * YAZICI <volkan.yazici@gmail.com>.
  *
  * metaphone()
  * -----------
@@ -43,188 +45,293 @@
  *
  */
 
-#include "fuzzystrmatch.h"
+#include "postgres.h"
+
+#include <ctype.h>
+
+#include "fmgr.h"
+#include "utils/builtins.h"
 
 PG_MODULE_MAGIC;
 
+
 /*
- * Calculates Levenshtein Distance between two strings.
- * Uses simplest and fastest cost model only, i.e. assumes a cost of 1 for
- * each deletion, substitution, or insertion.
+ * External declarations for exported functions
  */
+extern Datum levenshtein_with_costs(PG_FUNCTION_ARGS);
+extern Datum levenshtein(PG_FUNCTION_ARGS);
+extern Datum metaphone(PG_FUNCTION_ARGS);
+extern Datum soundex(PG_FUNCTION_ARGS);
+extern Datum difference(PG_FUNCTION_ARGS);
+
+/*
+ * Soundex
+ */
+static void _soundex(const char *instr, char *outstr);
+
+#define SOUNDEX_LEN 4
+
+/*									ABCDEFGHIJKLMNOPQRSTUVWXYZ */
+static const char *soundex_table = "01230120022455012623010202";
+
+static char
+soundex_code(char letter)
+{
+	letter = toupper((unsigned char) letter);
+	/* Defend against non-ASCII letters */
+	if (letter >= 'A' && letter <= 'Z')
+		return soundex_table[letter - 'A'];
+	return letter;
+}
+
+
+/*
+ * Levenshtein
+ */
+#define MAX_LEVENSHTEIN_STRLEN		255
+
+static int levenshtein_internal(const char *s, const char *t,
+					 int ins_c, int del_c, int sub_c);
+
+
+/*
+ * Metaphone
+ */
+#define MAX_METAPHONE_STRLEN		255
+
+/*
+ * Original code by Michael G Schwern starts here.
+ * Code slightly modified for use as PostgreSQL function.
+ */
+
+
+/**************************************************************************
+	metaphone -- Breaks english phrases down into their phonemes.
+
+	Input
+		word			--	An english word to be phonized
+		max_phonemes	--	How many phonemes to calculate.  If 0, then it
+							will phonize the entire phrase.
+		phoned_word		--	The final phonized word.  (We'll allocate the
+							memory.)
+	Output
+		error	--	A simple error flag, returns TRUE or FALSE
+
+	NOTES:	ALL non-alpha characters are ignored, this includes whitespace,
+	although non-alpha characters will break up phonemes.
+****************************************************************************/
+
+
+/**************************************************************************
+	my constants -- constants I like
+
+	Probably redundant.
+
+***************************************************************************/
+
+#define META_ERROR			FALSE
+#define META_SUCCESS		TRUE
+#define META_FAILURE		FALSE
+
+
+/*	I add modifications to the traditional metaphone algorithm that you
+	might find in books.  Define this if you want metaphone to behave
+	traditionally */
+#undef USE_TRADITIONAL_METAPHONE
+
+/* Special encodings */
+#define  SH		'X'
+#define  TH		'0'
+
+static char Lookahead(char *word, int how_far);
+static int	_metaphone(char *word, int max_phonemes, char **phoned_word);
+
+/* Metachar.h ... little bits about characters for metaphone */
+
+
+/*-- Character encoding array & accessing macros --*/
+/* Stolen directly out of the book... */
+static const char _codes[26] = {
+	1, 16, 4, 16, 9, 2, 4, 16, 9, 2, 0, 2, 2, 2, 1, 4, 0, 2, 4, 4, 1, 0, 0, 0, 8, 0
+/*	a  b c	d e f g  h i j k l m n o p q r s t u v w x y z */
+};
+
+static int
+getcode(char c)
+{
+	if (isalpha((unsigned char) c))
+	{
+		c = toupper((unsigned char) c);
+		/* Defend against non-ASCII letters */
+		if (c >= 'A' && c <= 'Z')
+			return _codes[c - 'A'];
+	}
+	return 0;
+}
+
+#define isvowel(c)	(getcode(c) & 1)	/* AEIOU */
+
+/* These letters are passed through unchanged */
+#define NOCHANGE(c) (getcode(c) & 2)	/* FJMNR */
+
+/* These form dipthongs when preceding H */
+#define AFFECTH(c)	(getcode(c) & 4)	/* CGPST */
+
+/* These make C and G soft */
+#define MAKESOFT(c) (getcode(c) & 8)	/* EIY */
+
+/* These prevent GH from becoming F */
+#define NOGHTOF(c)	(getcode(c) & 16)	/* BDH */
+
+
+/*
+ * levenshtein_internal - Calculates Levenshtein distance metric
+ *						  between supplied strings. Generally
+ *						  (1, 1, 1) penalty costs suffices common
+ *						  cases, but your mileage may vary.
+ */
+static int
+levenshtein_internal(const char *s, const char *t,
+					 int ins_c, int del_c, int sub_c)
+{
+	int			m,
+				n;
+	int		   *prev;
+	int		   *curr;
+	int			i,
+				j;
+	const char *x;
+	const char *y;
+
+	m = strlen(s);
+	n = strlen(t);
+
+	/*
+	 * If either m or n is 0, the answer is the other value. This makes sense
+	 * since it would take that many insertions to build a matching string
+	 */
+	if (!m)
+		return n;
+	if (!n)
+		return m;
+
+	/*
+	 * For security concerns, restrict excessive CPU+RAM usage. (This
+	 * implementation uses O(m) memory and has O(mn) complexity.)
+	 */
+	if (m > MAX_LEVENSHTEIN_STRLEN ||
+		n > MAX_LEVENSHTEIN_STRLEN)
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("argument exceeds the maximum length of %d bytes",
+						MAX_LEVENSHTEIN_STRLEN)));
+
+	/* One more cell for initialization column and row. */
+	++m;
+	++n;
+
+	/*
+	 * Instead of building an (m+1)x(n+1) array, we'll use two different
+	 * arrays of size m+1 for storing accumulated values. At each step one
+	 * represents the "previous" row and one is the "current" row of the
+	 * notional large array.
+	 */
+	prev = (int *) palloc(2 * m * sizeof(int));
+	curr = prev + m;
+
+	/* Initialize the "previous" row to 0..cols */
+	for (i = 0; i < m; i++)
+		prev[i] = i;
+
+	/* Loop through rows of the notional array */
+	for (y = t, j = 1; j < n; y++, j++)
+	{
+		int		   *temp;
+
+		/*
+		 * First cell must increment sequentially, as we're on the j'th row of
+		 * the (m+1)x(n+1) array.
+		 */
+		curr[0] = j;
+
+		for (x = s, i = 1; i < m; x++, i++)
+		{
+			int			ins;
+			int			del;
+			int			sub;
+
+			/* Calculate costs for probable operations. */
+			ins = prev[i] + ins_c;		/* Insertion	*/
+			del = curr[i - 1] + del_c;	/* Deletion		*/
+			sub = prev[i - 1] + ((*x == *y) ? 0 : sub_c);		/* Substitution */
+
+			/* Take the one with minimum cost. */
+			curr[i] = Min(ins, del);
+			curr[i] = Min(curr[i], sub);
+		}
+
+		/* Swap current row with previous row. */
+		temp = curr;
+		curr = prev;
+		prev = temp;
+	}
+
+	/*
+	 * Because the final value was swapped from the previous row to the
+	 * current row, that's where we'll find it.
+	 */
+	return prev[m - 1];
+}
+
+
+PG_FUNCTION_INFO_V1(levenshtein_with_costs);
+Datum
+levenshtein_with_costs(PG_FUNCTION_ARGS)
+{
+	char	   *src = TextDatumGetCString(PG_GETARG_DATUM(0));
+	char	   *dst = TextDatumGetCString(PG_GETARG_DATUM(1));
+	int			ins_c = PG_GETARG_INT32(2);
+	int			del_c = PG_GETARG_INT32(3);
+	int			sub_c = PG_GETARG_INT32(4);
+
+	PG_RETURN_INT32(levenshtein_internal(src, dst, ins_c, del_c, sub_c));
+}
+
+
 PG_FUNCTION_INFO_V1(levenshtein);
 Datum
 levenshtein(PG_FUNCTION_ARGS)
 {
-	char	   *str_s;
-	char	   *str_s0;
-	char	   *str_t;
-	int			cols = 0;
-	int			rows = 0;
-	int		   *u_cells;
-	int		   *l_cells;
-	int		   *tmp;
-	int			i;
-	int			j;
+	char	   *src = TextDatumGetCString(PG_GETARG_DATUM(0));
+	char	   *dst = TextDatumGetCString(PG_GETARG_DATUM(1));
 
-	/*
-	 * Fetch the arguments. str_s is referred to as the "source" cols = length
-	 * of source + 1 to allow for the initialization column str_t is referred
-	 * to as the "target", rows = length of target + 1 rows = length of target
-	 * + 1 to allow for the initialization row
-	 */
-	str_s = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(PG_GETARG_TEXT_P(0))));
-	str_t = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(PG_GETARG_TEXT_P(1))));
-
-	cols = strlen(str_s) + 1;
-	rows = strlen(str_t) + 1;
-
-	/*
-	 * Restrict the length of the strings being compared to something
-	 * reasonable because we will have to perform rows * cols calculations. If
-	 * longer strings need to be compared, increase MAX_LEVENSHTEIN_STRLEN to
-	 * suit (but within your tolerance for speed and memory usage).
-	 */
-	if ((cols > MAX_LEVENSHTEIN_STRLEN + 1) || (rows > MAX_LEVENSHTEIN_STRLEN + 1))
-		ereport(ERROR,
-				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("argument exceeds max length: %d",
-						MAX_LEVENSHTEIN_STRLEN)));
-
-	/*
-	 * If either rows or cols is 0, the answer is the other value. This makes
-	 * sense since it would take that many insertions the build a matching
-	 * string
-	 */
-
-	if (cols == 0)
-		PG_RETURN_INT32(rows);
-
-	if (rows == 0)
-		PG_RETURN_INT32(cols);
-
-	/*
-	 * Allocate two vectors of integers. One will be used for the "upper" row,
-	 * the other for the "lower" row. Initialize the "upper" row to 0..cols
-	 */
-	u_cells = palloc(sizeof(int) * cols);
-	for (i = 0; i < cols; i++)
-		u_cells[i] = i;
-
-	l_cells = palloc(sizeof(int) * cols);
-
-	/*
-	 * Use str_s0 to "rewind" the pointer to str_s in the nested for loop
-	 * below
-	 */
-	str_s0 = str_s;
-
-	/*
-	 * Loop through the rows, starting at row 1. Row 0 is used for the initial
-	 * "upper" row.
-	 */
-	for (j = 1; j < rows; j++)
-	{
-		/*
-		 * We'll always start with col 1, and initialize lower row col 0 to j
-		 */
-		l_cells[0] = j;
-
-		for (i = 1; i < cols; i++)
-		{
-			int			c = 0;
-			int			c1 = 0;
-			int			c2 = 0;
-			int			c3 = 0;
-
-			/*
-			 * The "cost" value is 0 if the character at the current col
-			 * position in the source string, matches the character at the
-			 * current row position in the target string; cost is 1 otherwise.
-			 */
-			c = (*str_s != *str_t);
-
-			/*
-			 * c1 is upper right cell plus 1
-			 */
-			c1 = u_cells[i] + 1;
-
-			/*
-			 * c2 is lower left cell plus 1
-			 */
-			c2 = l_cells[i - 1] + 1;
-
-			/*
-			 * c3 is cell diagonally above to the left plus "cost"
-			 */
-			c3 = u_cells[i - 1] + c;
-
-			/*
-			 * The lower right cell is set to the minimum of c1, c2, c3
-			 */
-			l_cells[i] = (c1 < c2 ? c1 : c2) < c3 ? (c1 < c2 ? c1 : c2) : c3;
-
-			/*
-			 * Increment the pointer to str_s
-			 */
-			str_s++;
-		}
-
-		/*
-		 * Lower row now becomes the upper row, and the upper row gets reused
-		 * as the new lower row.
-		 */
-		tmp = u_cells;
-		u_cells = l_cells;
-		l_cells = tmp;
-
-		/*
-		 * Increment the pointer to str_t
-		 */
-		str_t++;
-
-		/*
-		 * Rewind the pointer to str_s
-		 */
-		str_s = str_s0;
-	}
-
-	/*
-	 * Because the final value (at position row, col) was swapped from the
-	 * lower row to the upper row, that's where we'll find it.
-	 */
-	PG_RETURN_INT32(u_cells[cols - 1]);
+	PG_RETURN_INT32(levenshtein_internal(src, dst, 1, 1, 1));
 }
 
+
 /*
  * Calculates the metaphone of an input string.
  * Returns number of characters requested
  * (suggested value is 4)
  */
-#define GET_TEXT(cstrp) DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(cstrp)))
-
 PG_FUNCTION_INFO_V1(metaphone);
 Datum
 metaphone(PG_FUNCTION_ARGS)
 {
+	char	   *str_i = TextDatumGetCString(PG_GETARG_DATUM(0));
+	size_t		str_i_len = strlen(str_i);
 	int			reqlen;
-	char	   *str_i;
-	size_t		str_i_len;
 	char	   *metaph;
-	text	   *result_text;
 	int			retval;
 
-	str_i = DatumGetCString(DirectFunctionCall1(textout, PointerGetDatum(PG_GETARG_TEXT_P(0))));
-	str_i_len = strlen(str_i);
-
 	/* return an empty string if we receive one */
 	if (!(str_i_len > 0))
-		PG_RETURN_TEXT_P(GET_TEXT(""));
+		PG_RETURN_TEXT_P(cstring_to_text(""));
 
 	if (str_i_len > MAX_METAPHONE_STRLEN)
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("argument exceeds max length: %d",
+				 errmsg("argument exceeds the maximum length of %d bytes",
 						MAX_METAPHONE_STRLEN)));
 
 	if (!(str_i_len > 0))
@@ -236,7 +343,7 @@ metaphone(PG_FUNCTION_ARGS)
 	if (reqlen > MAX_METAPHONE_STRLEN)
 		ereport(ERROR,
 				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				 errmsg("output length exceeds max length: %d",
+				 errmsg("output exceeds the maximum length of %d bytes",
 						MAX_METAPHONE_STRLEN)));
 
 	if (!(reqlen > 0))
@@ -247,18 +354,12 @@ metaphone(PG_FUNCTION_ARGS)
 
 	retval = _metaphone(str_i, reqlen, &metaph);
 	if (retval == META_SUCCESS)
-	{
-		result_text = DatumGetTextP(DirectFunctionCall1(textin, CStringGetDatum(metaph)));
-		PG_RETURN_TEXT_P(result_text);
-	}
+		PG_RETURN_TEXT_P(cstring_to_text(metaph));
 	else
 	{
 		/* internal error */
 		elog(ERROR, "metaphone: failure");
-
-		/*
-		 * Keep the compiler quiet
-		 */
+		/* keep the compiler quiet */
 		PG_RETURN_NULL();
 	}
 }
@@ -267,9 +368,8 @@ metaphone(PG_FUNCTION_ARGS)
 /*
  * Original code by Michael G Schwern starts here.
  * Code slightly modified for use as PostgreSQL
- * function (palloc, etc). Original includes
- * are rolled into fuzzystrmatch.h
- *------------------------------------------------------------------*/
+ * function (palloc, etc).
+ */
 
 /* I suppose I could have been using a character pointer instead of
  * accessing the array directly... */
@@ -291,7 +391,7 @@ metaphone(PG_FUNCTION_ARGS)
 
 /* Allows us to safely look ahead an arbitrary # of letters */
 /* I probably could have just used strlen... */
-char
+static char
 Lookahead(char *word, int how_far)
 {
 	char		letter_ahead = '\0';	/* null by default */
@@ -317,14 +417,10 @@ Lookahead(char *word, int how_far)
 #define Isbreak(c)	(!isalpha((unsigned char) (c)))
 
 
-int
-_metaphone(
- /* IN */
-		   char *word,
+static int
+_metaphone(char *word,			/* IN */
 		   int max_phonemes,
- /* OUT */
-		   char **phoned_word
-)
+		   char **phoned_word)	/* OUT */
 {
 	int			w_idx = 0;		/* point in the phonization we're at. */
 	int			p_idx = 0;		/* end of the phoned phrase */
@@ -695,11 +791,11 @@ soundex(PG_FUNCTION_ARGS)
 	char		outstr[SOUNDEX_LEN + 1];
 	char	   *arg;
 
-	arg = _textout(PG_GETARG_TEXT_P(0));
+	arg = text_to_cstring(PG_GETARG_TEXT_P(0));
 
 	_soundex(arg, outstr);
 
-	PG_RETURN_TEXT_P(_textin(outstr));
+	PG_RETURN_TEXT_P(cstring_to_text(outstr));
 }
 
 static void
@@ -761,8 +857,8 @@ difference(PG_FUNCTION_ARGS)
 	int			i,
 				result;
 
-	_soundex(_textout(PG_GETARG_TEXT_P(0)), sndx1);
-	_soundex(_textout(PG_GETARG_TEXT_P(1)), sndx2);
+	_soundex(text_to_cstring(PG_GETARG_TEXT_P(0)), sndx1);
+	_soundex(text_to_cstring(PG_GETARG_TEXT_P(1)), sndx2);
 
 	result = 0;
 	for (i = 0; i < SOUNDEX_LEN; i++)
