@@ -1,14 +1,18 @@
 /*
+ * $PostgreSQL: pgsql/contrib/ltree/_ltree_gist.c,v 1.26 2009/06/11 14:48:51 momjian Exp $
+ *
+ *
  * GiST support for ltree[]
  * Teodor Sigaev <teodor@stack.net>
  */
+#include "postgres.h"
 
-#include "ltree.h"
 #include "access/gist.h"
 #include "access/skey.h"
 #include "utils/array.h"
-
 #include "crc32.h"
+#include "ltree.h"
+
 
 PG_FUNCTION_INFO_V1(_ltree_compress);
 Datum		_ltree_compress(PG_FUNCTION_ARGS);
@@ -55,7 +59,7 @@ static const uint8 number_of_ones[256] = {
 
 
 static void
-hashing(BITVECP sign, ltree * t)
+hashing(BITVECP sign, ltree *t)
 {
 	int			tlen = t->numlevel;
 	ltree_level *cur = LTREE_FIRST(t);
@@ -94,7 +98,7 @@ _ltree_compress(PG_FUNCTION_ARGS)
 					 errmsg("array must not contain nulls")));
 
 		key = (ltree_gist *) palloc(len);
-		key->len = len;
+		SET_VARSIZE(key, len);
 		key->flag = 0;
 
 		MemSet(LTG_SIGN(key), 0, ASIGLEN);
@@ -118,13 +122,14 @@ _ltree_compress(PG_FUNCTION_ARGS)
 
 		BITVECP		sign = LTG_SIGN(DatumGetPointer(entry->key));
 
-		ALOOPBYTE(
-				  if ((sign[i] & 0xff) != 0xff)
-				  PG_RETURN_POINTER(retval);
-		);
+		ALOOPBYTE
+		{
+			if ((sign[i] & 0xff) != 0xff)
+				PG_RETURN_POINTER(retval);
+		}
 		len = LTG_HDRSIZE;
 		key = (ltree_gist *) palloc(len);
-		key->len = len;
+		SET_VARSIZE(key, len);
 		key->flag = LTG_ALLTRUE;
 
 		retval = (GISTENTRY *) palloc(sizeof(GISTENTRY));
@@ -155,19 +160,20 @@ _ltree_same(PG_FUNCTION_ARGS)
 					sb = LTG_SIGN(b);
 
 		*result = true;
-		ALOOPBYTE(
-				  if (sa[i] != sb[i])
-				  {
-			*result = false;
-			break;
+		ALOOPBYTE
+		{
+			if (sa[i] != sb[i])
+			{
+				*result = false;
+				break;
+			}
 		}
-		);
 	}
 	PG_RETURN_POINTER(result);
 }
 
 static int4
-unionkey(BITVECP sbase, ltree_gist * add)
+unionkey(BITVECP sbase, ltree_gist *add)
 {
 	int4		i;
 	BITVECP		sadd = LTG_SIGN(add);
@@ -175,9 +181,8 @@ unionkey(BITVECP sbase, ltree_gist * add)
 	if (LTG_ISALLTRUE(add))
 		return 1;
 
-	ALOOPBYTE(
-			  sbase[i] |= sadd[i];
-	);
+	ALOOPBYTE
+		sbase[i] |= sadd[i];
 	return 0;
 }
 
@@ -204,10 +209,11 @@ _ltree_union(PG_FUNCTION_ARGS)
 
 	len = LTG_HDRSIZE + ((flag & LTG_ALLTRUE) ? 0 : ASIGLEN);
 	result = (ltree_gist *) palloc(len);
-	*size = result->len = len;
+	SET_VARSIZE(result, len);
 	result->flag = flag;
 	if (!LTG_ISALLTRUE(result))
 		memcpy((void *) LTG_SIGN(result), (void *) base, sizeof(ABITVEC));
+	*size = len;
 
 	PG_RETURN_POINTER(result);
 }
@@ -218,9 +224,8 @@ sizebitvec(BITVECP sign)
 	int4		size = 0,
 				i;
 
-	ALOOPBYTE(
-			  size += number_of_ones[(unsigned char) sign[i]];
-	);
+	ALOOPBYTE
+		size += number_of_ones[(unsigned char) sign[i]];
 	return size;
 }
 
@@ -231,15 +236,16 @@ hemdistsign(BITVECP a, BITVECP b)
 				diff,
 				dist = 0;
 
-	ALOOPBYTE(
-			  diff = (unsigned char) (a[i] ^ b[i]);
-	dist += number_of_ones[diff];
-	);
+	ALOOPBYTE
+	{
+		diff = (unsigned char) (a[i] ^ b[i]);
+		dist += number_of_ones[diff];
+	}
 	return dist;
 }
 
 static int
-hemdist(ltree_gist * a, ltree_gist * b)
+hemdist(ltree_gist *a, ltree_gist *b)
 {
 	if (LTG_ISALLTRUE(a))
 	{
@@ -270,7 +276,7 @@ typedef struct
 {
 	OffsetNumber pos;
 	int4		cost;
-}	SPLITCOST;
+} SPLITCOST;
 
 static int
 comparecost(const void *a, const void *b)
@@ -340,26 +346,26 @@ _ltree_picksplit(PG_FUNCTION_ARGS)
 	if (LTG_ISALLTRUE(GETENTRY(entryvec, seed_1)))
 	{
 		datum_l = (ltree_gist *) palloc(LTG_HDRSIZE);
-		datum_l->len = LTG_HDRSIZE;
+		SET_VARSIZE(datum_l, LTG_HDRSIZE);
 		datum_l->flag = LTG_ALLTRUE;
 	}
 	else
 	{
 		datum_l = (ltree_gist *) palloc(LTG_HDRSIZE + ASIGLEN);
-		datum_l->len = LTG_HDRSIZE + ASIGLEN;
+		SET_VARSIZE(datum_l, LTG_HDRSIZE + ASIGLEN);
 		datum_l->flag = 0;
 		memcpy((void *) LTG_SIGN(datum_l), (void *) LTG_SIGN(GETENTRY(entryvec, seed_1)), sizeof(ABITVEC));
 	}
 	if (LTG_ISALLTRUE(GETENTRY(entryvec, seed_2)))
 	{
 		datum_r = (ltree_gist *) palloc(LTG_HDRSIZE);
-		datum_r->len = LTG_HDRSIZE;
+		SET_VARSIZE(datum_r, LTG_HDRSIZE);
 		datum_r->flag = LTG_ALLTRUE;
 	}
 	else
 	{
 		datum_r = (ltree_gist *) palloc(LTG_HDRSIZE + ASIGLEN);
-		datum_r->len = LTG_HDRSIZE + ASIGLEN;
+		SET_VARSIZE(datum_r, LTG_HDRSIZE + ASIGLEN);
 		datum_r->flag = 0;
 		memcpy((void *) LTG_SIGN(datum_r), (void *) LTG_SIGN(GETENTRY(entryvec, seed_2)), sizeof(ABITVEC));
 	}
@@ -409,9 +415,8 @@ _ltree_picksplit(PG_FUNCTION_ARGS)
 			else
 			{
 				ptr = LTG_SIGN(_j);
-				ALOOPBYTE(
-						  union_l[i] |= ptr[i];
-				);
+				ALOOPBYTE
+					union_l[i] |= ptr[i];
 			}
 			*left++ = j;
 			v->spl_nleft++;
@@ -426,9 +431,8 @@ _ltree_picksplit(PG_FUNCTION_ARGS)
 			else
 			{
 				ptr = LTG_SIGN(_j);
-				ALOOPBYTE(
-						  union_r[i] |= ptr[i];
-				);
+				ALOOPBYTE
+					union_r[i] |= ptr[i];
 			}
 			*right++ = j;
 			v->spl_nright++;
@@ -444,7 +448,7 @@ _ltree_picksplit(PG_FUNCTION_ARGS)
 }
 
 static bool
-gist_te(ltree_gist * key, ltree * query)
+gist_te(ltree_gist *key, ltree *query)
 {
 	ltree_level *curq = LTREE_FIRST(query);
 	BITVECP		sign = LTG_SIGN(key);
@@ -467,13 +471,13 @@ gist_te(ltree_gist * key, ltree * query)
 }
 
 static bool
-checkcondition_bit(void *checkval, ITEM * val)
+checkcondition_bit(void *checkval, ITEM *val)
 {
 	return (FLG_CANLOOKSIGN(val->flag)) ? GETBIT(checkval, AHASHVAL(val->val)) : true;
 }
 
 static bool
-gist_qtxt(ltree_gist * key, ltxtquery * query)
+gist_qtxt(ltree_gist *key, ltxtquery *query)
 {
 	if (LTG_ISALLTRUE(key))
 		return true;
@@ -486,7 +490,7 @@ gist_qtxt(ltree_gist * key, ltxtquery * query)
 }
 
 static bool
-gist_qe(ltree_gist * key, lquery * query)
+gist_qe(ltree_gist *key, lquery *query)
 {
 	lquery_level *curq = LQUERY_FIRST(query);
 	BITVECP		sign = LTG_SIGN(key);
@@ -525,7 +529,7 @@ gist_qe(ltree_gist * key, lquery * query)
 }
 
 static bool
-_arrq_cons(ltree_gist * key, ArrayType *_query)
+_arrq_cons(ltree_gist *key, ArrayType *_query)
 {
 	lquery	   *query = (lquery *) ARR_DATA_PTR(_query);
 	int			num = ArrayGetNItems(ARR_NDIM(_query), ARR_DIMS(_query));
@@ -554,10 +558,16 @@ _ltree_consistent(PG_FUNCTION_ARGS)
 {
 	GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
 	char	   *query = (char *) DatumGetPointer(PG_DETOAST_DATUM(PG_GETARG_DATUM(1)));
-	ltree_gist *key = (ltree_gist *) DatumGetPointer(entry->key);
 	StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);
+
+	/* Oid		subtype = PG_GETARG_OID(3); */
+	bool	   *recheck = (bool *) PG_GETARG_POINTER(4);
+	ltree_gist *key = (ltree_gist *) DatumGetPointer(entry->key);
 	bool		res = false;
 
+	/* All cases served by this function are inexact */
+	*recheck = true;
+
 	switch (strategy)
 	{
 		case 10:
