@@ -4,10 +4,10 @@
  *	  Two-phase-commit resource managers definition
  *
  *
- * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
- * $PostgreSQL: pgsql/src/include/access/twophase_rmgr.h,v 1.2.2.1 2009/11/23 09:59:21 heikki Exp $
+ * $PostgreSQL: pgsql/src/include/access/twophase_rmgr.h,v 1.7 2009/01/01 17:23:56 momjian Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -26,8 +26,8 @@ typedef uint8 TwoPhaseRmgrId;
 #define TWOPHASE_RM_INVAL_ID		2
 #define TWOPHASE_RM_FLATFILES_ID	3
 #define TWOPHASE_RM_NOTIFY_ID		4
-#define TWOPHASE_RM_MULTIXACT_ID	5
-#define TWOPHASE_RM_MAX_ID			TWOPHASE_RM_MULTIXACT_ID
+#define TWOPHASE_RM_PGSTAT_ID		5
+#define TWOPHASE_RM_MAX_ID			TWOPHASE_RM_PGSTAT_ID
 
 extern const TwoPhaseCallback twophase_recover_callbacks[];
 extern const TwoPhaseCallback twophase_postcommit_callbacks[];
