@@ -1,42 +1,19 @@
-/**********************************************************************
- * pl_funcs.c		- Misc functins for the PL/pgSQL
+/*-------------------------------------------------------------------------
+ *
+ * pl_funcs.c		- Misc functions for the PL/pgSQL
  *			  procedural language
  *
+ * Portions Copyright (c) 1996-2009, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1994, Regents of the University of California
+ *
+ *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/pl/plpgsql/src/pl_funcs.c,v 1.30.2.1 2004/02/21 00:35:13 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/pl/plpgsql/src/pl_funcs.c,v 1.79 2009/06/11 14:49:14 momjian Exp $
  *
- *	  This software is copyrighted by Jan Wieck - Hamburg.
- *
- *	  The author hereby grants permission  to  use,  copy,	modify,
- *	  distribute,  and	license this software and its documentation
- *	  for any purpose, provided that existing copyright notices are
- *	  retained	in	all  copies  and  that	this notice is included
- *	  verbatim in any distributions. No written agreement, license,
- *	  or  royalty  fee	is required for any of the authorized uses.
- *	  Modifications to this software may be  copyrighted  by  their
- *	  author  and  need  not  follow  the licensing terms described
- *	  here, provided that the new terms are  clearly  indicated  on
- *	  the first page of each file where they apply.
- *
- *	  IN NO EVENT SHALL THE AUTHOR OR DISTRIBUTORS BE LIABLE TO ANY
- *	  PARTY  FOR  DIRECT,	INDIRECT,	SPECIAL,   INCIDENTAL,	 OR
- *	  CONSEQUENTIAL   DAMAGES  ARISING	OUT  OF  THE  USE  OF  THIS
- *	  SOFTWARE, ITS DOCUMENTATION, OR ANY DERIVATIVES THEREOF, EVEN
- *	  IF  THE  AUTHOR  HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH
- *	  DAMAGE.
- *
- *	  THE  AUTHOR  AND	DISTRIBUTORS  SPECIFICALLY	 DISCLAIM	ANY
- *	  WARRANTIES,  INCLUDING,  BUT	NOT  LIMITED  TO,  THE	IMPLIED
- *	  WARRANTIES  OF  MERCHANTABILITY,	FITNESS  FOR  A  PARTICULAR
- *	  PURPOSE,	AND NON-INFRINGEMENT.  THIS SOFTWARE IS PROVIDED ON
- *	  AN "AS IS" BASIS, AND THE AUTHOR	AND  DISTRIBUTORS  HAVE  NO
- *	  OBLIGATION   TO	PROVIDE   MAINTENANCE,	 SUPPORT,  UPDATES,
- *	  ENHANCEMENTS, OR MODIFICATIONS.
- *
- **********************************************************************/
+ *-------------------------------------------------------------------------
+ */
 
 #include "plpgsql.h"
-#include "pl.tab.h"
 
 #include <ctype.h>
 
@@ -56,10 +33,10 @@ static bool ns_localmode = false;
  * ----------
  */
 void
-plpgsql_dstring_init(PLpgSQL_dstring * ds)
+plpgsql_dstring_init(PLpgSQL_dstring *ds)
 {
 	ds->value = palloc(ds->alloc = 512);
-	ds->used = 0;
+	ds->used = 1;
 	ds->value[0] = '\0';
 }
 
@@ -69,34 +46,58 @@ plpgsql_dstring_init(PLpgSQL_dstring * ds)
  * ----------
  */
 void
-plpgsql_dstring_free(PLpgSQL_dstring * ds)
+plpgsql_dstring_free(PLpgSQL_dstring *ds)
 {
 	pfree(ds->value);
 }
 
+static void
+plpgsql_dstring_expand(PLpgSQL_dstring *ds, int needed)
+{
+	/* Don't allow truncating the string */
+	Assert(needed > ds->alloc);
+	Assert(ds->used <= ds->alloc);
+
+	/* Might have to double more than once, if needed is large */
+	do
+	{
+		ds->alloc *= 2;
+	} while (needed > ds->alloc);
+	ds->value = repalloc(ds->value, ds->alloc);
+}
 
 /* ----------
  * plpgsql_dstring_append		Dynamic string extending
  * ----------
  */
 void
-plpgsql_dstring_append(PLpgSQL_dstring * ds, char *str)
+plpgsql_dstring_append(PLpgSQL_dstring *ds, const char *str)
 {
 	int			len = strlen(str);
-	int			needed = ds->used + len + 1;
+	int			needed = ds->used + len;
 
 	if (needed > ds->alloc)
-	{
-		/* might have to double more than once, if len is large */
-		do
-		{
-			ds->alloc *= 2;
-		} while (needed > ds->alloc);
-		ds->value = repalloc(ds->value, ds->alloc);
-	}
+		plpgsql_dstring_expand(ds, needed);
 
-	strcpy(&(ds->value[ds->used]), str);
+	memcpy(&(ds->value[ds->used - 1]), str, len);
 	ds->used += len;
+	ds->value[ds->used - 1] = '\0';
+}
+
+/* ----------
+ * plpgsql_dstring_append_char	Append a single character
+ *								to a dynamic string
+ * ----------
+ */
+void
+plpgsql_dstring_append_char(PLpgSQL_dstring *ds, char c)
+{
+	if (ds->used == ds->alloc)
+		plpgsql_dstring_expand(ds, ds->used + 1);
+
+	ds->value[ds->used - 1] = c;
+	ds->value[ds->used] = '\0';
+	ds->used++;
 }
 
 
@@ -105,7 +106,7 @@ plpgsql_dstring_append(PLpgSQL_dstring * ds, char *str)
  * ----------
  */
 char *
-plpgsql_dstring_get(PLpgSQL_dstring * ds)
+plpgsql_dstring_get(PLpgSQL_dstring *ds)
 {
 	return ds->value;
 }
@@ -124,9 +125,15 @@ plpgsql_ns_init(void)
 
 
 /* ----------
- * plpgsql_ns_setlocal			Tell plpgsql_ns_lookup to or to
- *					not look into the current level
- *					only.
+ * plpgsql_ns_setlocal			Tell plpgsql_ns_lookup whether to
+ *					look into the current level only.
+ *
+ * This is a crock, but in the current design we need it because scan.l
+ * initiates name lookup, and the scanner does not know whether we are
+ * examining a name being declared in a DECLARE section.  For that case
+ * we only want to know if there is a conflicting name earlier in the
+ * same DECLARE section.  So the grammar must temporarily set local mode
+ * before scanning decl_varnames.
  * ----------
  */
 bool
@@ -145,15 +152,14 @@ plpgsql_ns_setlocal(bool flag)
  * ----------
  */
 void
-plpgsql_ns_push(char *label)
+plpgsql_ns_push(const char *label)
 {
 	PLpgSQL_ns *new;
 
 	if (label == NULL)
 		label = "";
 
-	new = palloc(sizeof(PLpgSQL_ns));
-	memset(new, 0, sizeof(PLpgSQL_ns));
+	new = palloc0(sizeof(PLpgSQL_ns));
 	new->upper = ns_current;
 	ns_current = new;
 
@@ -166,7 +172,7 @@ plpgsql_ns_push(char *label)
  * ----------
  */
 void
-plpgsql_ns_pop()
+plpgsql_ns_pop(void)
 {
 	int			i;
 	PLpgSQL_ns *old;
@@ -187,7 +193,7 @@ plpgsql_ns_pop()
  * ----------
  */
 void
-plpgsql_ns_additem(int itemtype, int itemno, char *name)
+plpgsql_ns_additem(int itemtype, int itemno, const char *name)
 {
 	PLpgSQL_ns *ns = ns_current;
 	PLpgSQL_nsitem *nse;
@@ -205,7 +211,7 @@ plpgsql_ns_additem(int itemtype, int itemno, char *name)
 		{
 			ns->items_alloc *= 2;
 			ns->items = repalloc(ns->items,
-							 sizeof(PLpgSQL_nsitem *) * ns->items_alloc);
+								 sizeof(PLpgSQL_nsitem *) * ns->items_alloc);
 		}
 	}
 
@@ -218,59 +224,98 @@ plpgsql_ns_additem(int itemtype, int itemno, char *name)
 
 
 /* ----------
- * plpgsql_ns_lookup			Lookup for a word in the namestack
+ * plpgsql_ns_lookup			Lookup an identifier in the namestack
+ *
+ * Note that this only searches for variables, not labels.
+ *
+ * name1 must be non-NULL.	Pass NULL for name2 and/or name3 if parsing a name
+ * with fewer than three components.
+ *
+ * If names_used isn't NULL, *names_used receives the number of names
+ * matched: 0 if no match, 1 if name1 matched an unqualified variable name,
+ * 2 if name1 and name2 matched a block label + variable name.
+ *
+ * Note that name3 is never directly matched to anything.  However, if it
+ * isn't NULL, we will disregard qualified matches to scalar variables.
+ * Similarly, if name2 isn't NULL, we disregard unqualified matches to
+ * scalar variables.
  * ----------
  */
 PLpgSQL_nsitem *
-plpgsql_ns_lookup(char *name, char *label)
+plpgsql_ns_lookup(const char *name1, const char *name2, const char *name3,
+				  int *names_used)
 {
 	PLpgSQL_ns *ns;
 	int			i;
 
-	/*
-	 * If a label is specified, lookup only in that
-	 */
-	if (label != NULL)
-	{
-		for (ns = ns_current; ns != NULL; ns = ns->upper)
-		{
-			if (!strcmp(ns->items[0]->name, label))
-			{
-				for (i = 1; i < ns->items_used; i++)
-				{
-					if (!strcmp(ns->items[i]->name, name))
-						return ns->items[i];
-				}
-				return NULL;	/* name not found in specified label */
-			}
-		}
-		return NULL;			/* label not found */
-	}
-
-	/*
-	 * No label given, lookup for visible labels ignoring localmode
-	 */
+	/* Scan each level of the namestack */
 	for (ns = ns_current; ns != NULL; ns = ns->upper)
 	{
-		if (!strcmp(ns->items[0]->name, name))
+		/* Check for unqualified match to variable name */
+		for (i = 1; i < ns->items_used; i++)
+		{
+			PLpgSQL_nsitem *nsitem = ns->items[i];
+
+			if (strcmp(nsitem->name, name1) == 0)
+			{
+				if (name2 == NULL ||
+					nsitem->itemtype != PLPGSQL_NSTYPE_VAR)
+				{
+					if (names_used)
+						*names_used = 1;
+					return nsitem;
+				}
+			}
+		}
+
+		/* Check for qualified match to variable name */
+		if (name2 != NULL &&
+			strcmp(ns->items[0]->name, name1) == 0)
+		{
+			for (i = 1; i < ns->items_used; i++)
+			{
+				PLpgSQL_nsitem *nsitem = ns->items[i];
+
+				if (strcmp(nsitem->name, name2) == 0)
+				{
+					if (name3 == NULL ||
+						nsitem->itemtype != PLPGSQL_NSTYPE_VAR)
+					{
+						if (names_used)
+							*names_used = 2;
+						return nsitem;
+					}
+				}
+			}
+		}
+
+		if (ns_localmode)
+			break;				/* do not look into upper levels */
+	}
+
+	/* This is just to suppress possibly-uninitialized-variable warnings */
+	if (names_used)
+		*names_used = 0;
+	return NULL;				/* No match found */
+}
+
+
+/* ----------
+ * plpgsql_ns_lookup_label		Lookup a label in the namestack
+ * ----------
+ */
+PLpgSQL_nsitem *
+plpgsql_ns_lookup_label(const char *name)
+{
+	PLpgSQL_ns *ns;
+
+	for (ns = ns_current; ns != NULL; ns = ns->upper)
+	{
+		if (strcmp(ns->items[0]->name, name) == 0)
 			return ns->items[0];
 	}
 
-	/*
-	 * Finally lookup name in the namestack
-	 */
-	for (ns = ns_current; ns != NULL; ns = ns->upper)
-	{
-		for (i = 1; i < ns->items_used; i++)
-		{
-			if (!strcmp(ns->items[i]->name, name))
-				return ns->items[i];
-		}
-		if (ns_localmode)
-			return NULL;		/* name not found in current namespace */
-	}
-
-	return NULL;
+	return NULL;				/* label not found */
 }
 
 
@@ -285,10 +330,6 @@ plpgsql_ns_rename(char *oldname, char *newname)
 	PLpgSQL_nsitem *newitem;
 	int			i;
 
-	/*
-	 * Lookup in the current namespace only
-	 */
-
 	/*
 	 * Lookup name in the namestack
 	 */
@@ -296,7 +337,7 @@ plpgsql_ns_rename(char *oldname, char *newname)
 	{
 		for (i = 1; i < ns->items_used; i++)
 		{
-			if (!strcmp(ns->items[i]->name, oldname))
+			if (strcmp(ns->items[i]->name, oldname) == 0)
 			{
 				newitem = palloc(sizeof(PLpgSQL_nsitem) + strlen(newname));
 				newitem->itemtype = ns->items[i]->itemtype;
@@ -315,7 +356,7 @@ plpgsql_ns_rename(char *oldname, char *newname)
 
 	ereport(ERROR,
 			(errcode(ERRCODE_UNDEFINED_OBJECT),
-			 errmsg("there is no variable \"%s\" in the current block",
+			 errmsg("variable \"%s\" does not exist in the current block",
 					oldname)));
 }
 
@@ -355,7 +396,7 @@ plpgsql_convert_ident(const char *s, char **output, int numidents)
 		{
 			/* Quoted identifier: copy, collapsing out doubled quotes */
 
-			curident = palloc(strlen(s) + 1); /* surely enough room */
+			curident = palloc(strlen(s) + 1);	/* surely enough room */
 			cp = curident;
 			s++;
 			while (*s)
@@ -371,21 +412,21 @@ plpgsql_convert_ident(const char *s, char **output, int numidents)
 			if (*s != '"')		/* should not happen if lexer checked */
 				ereport(ERROR,
 						(errcode(ERRCODE_SYNTAX_ERROR),
-						 errmsg("unterminated \" in name: %s", sstart)));
+					   errmsg("unterminated \" in identifier: %s", sstart)));
 			s++;
 			*cp = '\0';
 			/* Truncate to NAMEDATALEN */
-			truncate_identifier(curident, cp-curident, false);
+			truncate_identifier(curident, cp - curident, false);
 		}
 		else
 		{
 			/* Normal identifier: extends till dot or whitespace */
 			const char *thisstart = s;
 
-			while (*s && *s != '.' && !isspace((unsigned char) *s))
+			while (*s && *s != '.' && !scanner_isspace(*s))
 				s++;
 			/* Downcase and truncate to NAMEDATALEN */
-			curident = downcase_truncate_identifier(thisstart, s-thisstart,
+			curident = downcase_truncate_identifier(thisstart, s - thisstart,
 													false);
 		}
 
@@ -395,17 +436,17 @@ plpgsql_convert_ident(const char *s, char **output, int numidents)
 		else
 			ereport(ERROR,
 					(errcode(ERRCODE_SYNTAX_ERROR),
-				   errmsg("qualified identifier cannot be used here: %s",
-						  sstart)));
+					 errmsg("qualified identifier cannot be used here: %s",
+							sstart)));
 
 		/* If not done, skip whitespace, dot, whitespace */
 		if (*s)
 		{
-			while (*s && isspace((unsigned char) *s))
+			while (*s && scanner_isspace(*s))
 				s++;
 			if (*s++ != '.')
 				elog(ERROR, "expected dot between identifiers: %s", sstart);
-			while (*s && isspace((unsigned char) *s))
+			while (*s && scanner_isspace(*s))
 				s++;
 			if (*s == '\0')
 				elog(ERROR, "expected another identifier: %s", sstart);
@@ -422,50 +463,54 @@ plpgsql_convert_ident(const char *s, char **output, int numidents)
  * Statement type as a string, for use in error messages etc.
  */
 const char *
-plpgsql_stmt_typename(PLpgSQL_stmt * stmt)
+plpgsql_stmt_typename(PLpgSQL_stmt *stmt)
 {
-	switch (stmt->cmd_type)
+	switch ((enum PLpgSQL_stmt_types) stmt->cmd_type)
 	{
 		case PLPGSQL_STMT_BLOCK:
-			return "block variables initialization";
+			return _("statement block");
 		case PLPGSQL_STMT_ASSIGN:
-			return "assignment";
+			return _("assignment");
 		case PLPGSQL_STMT_IF:
-			return "if";
+			return "IF";
+		case PLPGSQL_STMT_CASE:
+			return "CASE";
 		case PLPGSQL_STMT_LOOP:
-			return "loop";
+			return "LOOP";
 		case PLPGSQL_STMT_WHILE:
-			return "while";
+			return "WHILE";
 		case PLPGSQL_STMT_FORI:
-			return "for with integer loopvar";
+			return _("FOR with integer loop variable");
 		case PLPGSQL_STMT_FORS:
-			return "for over select rows";
-		case PLPGSQL_STMT_SELECT:
-			return "select into variables";
+			return _("FOR over SELECT rows");
+		case PLPGSQL_STMT_FORC:
+			return _("FOR over cursor");
 		case PLPGSQL_STMT_EXIT:
-			return "exit";
+			return "EXIT";
 		case PLPGSQL_STMT_RETURN:
-			return "return";
+			return "RETURN";
 		case PLPGSQL_STMT_RETURN_NEXT:
-			return "return next";
+			return "RETURN NEXT";
+		case PLPGSQL_STMT_RETURN_QUERY:
+			return "RETURN QUERY";
 		case PLPGSQL_STMT_RAISE:
-			return "raise";
+			return "RAISE";
 		case PLPGSQL_STMT_EXECSQL:
-			return "SQL statement";
+			return _("SQL statement");
 		case PLPGSQL_STMT_DYNEXECUTE:
-			return "execute statement";
+			return _("EXECUTE statement");
 		case PLPGSQL_STMT_DYNFORS:
-			return "for over execute statement";
+			return _("FOR over EXECUTE statement");
 		case PLPGSQL_STMT_GETDIAG:
-			return "get diagnostics";
+			return "GET DIAGNOSTICS";
 		case PLPGSQL_STMT_OPEN:
-			return "open";
+			return "OPEN";
 		case PLPGSQL_STMT_FETCH:
-			return "fetch";
+			return "FETCH";
 		case PLPGSQL_STMT_CLOSE:
-			return "close";
+			return "CLOSE";
 		case PLPGSQL_STMT_PERFORM:
-			return "perform";
+			return "PERFORM";
 	}
 
 	return "unknown";
@@ -477,33 +522,36 @@ plpgsql_stmt_typename(PLpgSQL_stmt * stmt)
  **********************************************************************/
 static int	dump_indent;
 
-static void dump_ind();
-static void dump_stmt(PLpgSQL_stmt * stmt);
-static void dump_block(PLpgSQL_stmt_block * block);
-static void dump_assign(PLpgSQL_stmt_assign * stmt);
-static void dump_if(PLpgSQL_stmt_if * stmt);
-static void dump_loop(PLpgSQL_stmt_loop * stmt);
-static void dump_while(PLpgSQL_stmt_while * stmt);
-static void dump_fori(PLpgSQL_stmt_fori * stmt);
-static void dump_fors(PLpgSQL_stmt_fors * stmt);
-static void dump_select(PLpgSQL_stmt_select * stmt);
-static void dump_exit(PLpgSQL_stmt_exit * stmt);
-static void dump_return(PLpgSQL_stmt_return * stmt);
-static void dump_return_next(PLpgSQL_stmt_return_next * stmt);
-static void dump_raise(PLpgSQL_stmt_raise * stmt);
-static void dump_execsql(PLpgSQL_stmt_execsql * stmt);
-static void dump_dynexecute(PLpgSQL_stmt_dynexecute * stmt);
-static void dump_dynfors(PLpgSQL_stmt_dynfors * stmt);
-static void dump_getdiag(PLpgSQL_stmt_getdiag * stmt);
-static void dump_open(PLpgSQL_stmt_open * stmt);
-static void dump_fetch(PLpgSQL_stmt_fetch * stmt);
-static void dump_close(PLpgSQL_stmt_close * stmt);
-static void dump_perform(PLpgSQL_stmt_perform * stmt);
-static void dump_expr(PLpgSQL_expr * expr);
+static void dump_ind(void);
+static void dump_stmt(PLpgSQL_stmt *stmt);
+static void dump_block(PLpgSQL_stmt_block *block);
+static void dump_assign(PLpgSQL_stmt_assign *stmt);
+static void dump_if(PLpgSQL_stmt_if *stmt);
+static void dump_case(PLpgSQL_stmt_case *stmt);
+static void dump_loop(PLpgSQL_stmt_loop *stmt);
+static void dump_while(PLpgSQL_stmt_while *stmt);
+static void dump_fori(PLpgSQL_stmt_fori *stmt);
+static void dump_fors(PLpgSQL_stmt_fors *stmt);
+static void dump_forc(PLpgSQL_stmt_forc *stmt);
+static void dump_exit(PLpgSQL_stmt_exit *stmt);
+static void dump_return(PLpgSQL_stmt_return *stmt);
+static void dump_return_next(PLpgSQL_stmt_return_next *stmt);
+static void dump_return_query(PLpgSQL_stmt_return_query *stmt);
+static void dump_raise(PLpgSQL_stmt_raise *stmt);
+static void dump_execsql(PLpgSQL_stmt_execsql *stmt);
+static void dump_dynexecute(PLpgSQL_stmt_dynexecute *stmt);
+static void dump_dynfors(PLpgSQL_stmt_dynfors *stmt);
+static void dump_getdiag(PLpgSQL_stmt_getdiag *stmt);
+static void dump_open(PLpgSQL_stmt_open *stmt);
+static void dump_fetch(PLpgSQL_stmt_fetch *stmt);
+static void dump_cursor_direction(PLpgSQL_stmt_fetch *stmt);
+static void dump_close(PLpgSQL_stmt_close *stmt);
+static void dump_perform(PLpgSQL_stmt_perform *stmt);
+static void dump_expr(PLpgSQL_expr *expr);
 
 
 static void
-dump_ind()
+dump_ind(void)
 {
 	int			i;
 
@@ -512,10 +560,10 @@ dump_ind()
 }
 
 static void
-dump_stmt(PLpgSQL_stmt * stmt)
+dump_stmt(PLpgSQL_stmt *stmt)
 {
 	printf("%3d:", stmt->lineno);
-	switch (stmt->cmd_type)
+	switch ((enum PLpgSQL_stmt_types) stmt->cmd_type)
 	{
 		case PLPGSQL_STMT_BLOCK:
 			dump_block((PLpgSQL_stmt_block *) stmt);
@@ -526,6 +574,9 @@ dump_stmt(PLpgSQL_stmt * stmt)
 		case PLPGSQL_STMT_IF:
 			dump_if((PLpgSQL_stmt_if *) stmt);
 			break;
+		case PLPGSQL_STMT_CASE:
+			dump_case((PLpgSQL_stmt_case *) stmt);
+			break;
 		case PLPGSQL_STMT_LOOP:
 			dump_loop((PLpgSQL_stmt_loop *) stmt);
 			break;
@@ -538,8 +589,8 @@ dump_stmt(PLpgSQL_stmt * stmt)
 		case PLPGSQL_STMT_FORS:
 			dump_fors((PLpgSQL_stmt_fors *) stmt);
 			break;
-		case PLPGSQL_STMT_SELECT:
-			dump_select((PLpgSQL_stmt_select *) stmt);
+		case PLPGSQL_STMT_FORC:
+			dump_forc((PLpgSQL_stmt_forc *) stmt);
 			break;
 		case PLPGSQL_STMT_EXIT:
 			dump_exit((PLpgSQL_stmt_exit *) stmt);
@@ -550,6 +601,9 @@ dump_stmt(PLpgSQL_stmt * stmt)
 		case PLPGSQL_STMT_RETURN_NEXT:
 			dump_return_next((PLpgSQL_stmt_return_next *) stmt);
 			break;
+		case PLPGSQL_STMT_RETURN_QUERY:
+			dump_return_query((PLpgSQL_stmt_return_query *) stmt);
+			break;
 		case PLPGSQL_STMT_RAISE:
 			dump_raise((PLpgSQL_stmt_raise *) stmt);
 			break;
@@ -584,9 +638,19 @@ dump_stmt(PLpgSQL_stmt * stmt)
 }
 
 static void
-dump_block(PLpgSQL_stmt_block * block)
+dump_stmts(List *stmts)
+{
+	ListCell   *s;
+
+	dump_indent += 2;
+	foreach(s, stmts)
+		dump_stmt((PLpgSQL_stmt *) lfirst(s));
+	dump_indent -= 2;
+}
+
+static void
+dump_block(PLpgSQL_stmt_block *block)
 {
-	int			i;
 	char	   *name;
 
 	if (block->label == NULL)
@@ -597,17 +661,36 @@ dump_block(PLpgSQL_stmt_block * block)
 	dump_ind();
 	printf("BLOCK <<%s>>\n", name);
 
-	dump_indent += 2;
-	for (i = 0; i < block->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (block->body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(block->body);
+
+	if (block->exceptions)
+	{
+		ListCell   *e;
+
+		foreach(e, block->exceptions->exc_list)
+		{
+			PLpgSQL_exception *exc = (PLpgSQL_exception *) lfirst(e);
+			PLpgSQL_condition *cond;
+
+			dump_ind();
+			printf("    EXCEPTION WHEN ");
+			for (cond = exc->conditions; cond; cond = cond->next)
+			{
+				if (cond != exc->conditions)
+					printf(" OR ");
+				printf("%s", cond->condname);
+			}
+			printf(" THEN\n");
+			dump_stmts(exc->action);
+		}
+	}
 
 	dump_ind();
 	printf("    END -- %s\n", name);
 }
 
 static void
-dump_assign(PLpgSQL_stmt_assign * stmt)
+dump_assign(PLpgSQL_stmt_assign *stmt)
 {
 	dump_ind();
 	printf("ASSIGN var %d := ", stmt->varno);
@@ -616,73 +699,93 @@ dump_assign(PLpgSQL_stmt_assign * stmt)
 }
 
 static void
-dump_if(PLpgSQL_stmt_if * stmt)
+dump_if(PLpgSQL_stmt_if *stmt)
 {
-	int			i;
-
 	dump_ind();
 	printf("IF ");
 	dump_expr(stmt->cond);
 	printf(" THEN\n");
 
-	dump_indent += 2;
-	for (i = 0; i < stmt->true_body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->true_body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(stmt->true_body);
 
-	dump_ind();
-	printf("    ELSE\n");
-
-	dump_indent += 2;
-	for (i = 0; i < stmt->false_body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->false_body->stmts[i]));
-	dump_indent -= 2;
+	if (stmt->false_body != NIL)
+	{
+		dump_ind();
+		printf("    ELSE\n");
+		dump_stmts(stmt->false_body);
+	}
 
 	dump_ind();
 	printf("    ENDIF\n");
 }
 
 static void
-dump_loop(PLpgSQL_stmt_loop * stmt)
+dump_case(PLpgSQL_stmt_case *stmt)
 {
-	int			i;
+	ListCell   *l;
 
+	dump_ind();
+	printf("CASE %d ", stmt->t_varno);
+	if (stmt->t_expr)
+		dump_expr(stmt->t_expr);
+	printf("\n");
+	dump_indent += 6;
+	foreach(l, stmt->case_when_list)
+	{
+		PLpgSQL_case_when *cwt = (PLpgSQL_case_when *) lfirst(l);
+
+		dump_ind();
+		printf("WHEN ");
+		dump_expr(cwt->expr);
+		printf("\n");
+		dump_ind();
+		printf("THEN\n");
+		dump_indent += 2;
+		dump_stmts(cwt->stmts);
+		dump_indent -= 2;
+	}
+	if (stmt->have_else)
+	{
+		dump_ind();
+		printf("ELSE\n");
+		dump_indent += 2;
+		dump_stmts(stmt->else_stmts);
+		dump_indent -= 2;
+	}
+	dump_indent -= 6;
+	dump_ind();
+	printf("    ENDCASE\n");
+}
+
+static void
+dump_loop(PLpgSQL_stmt_loop *stmt)
+{
 	dump_ind();
 	printf("LOOP\n");
 
-	dump_indent += 2;
-	for (i = 0; i < stmt->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(stmt->body);
 
 	dump_ind();
 	printf("    ENDLOOP\n");
 }
 
 static void
-dump_while(PLpgSQL_stmt_while * stmt)
+dump_while(PLpgSQL_stmt_while *stmt)
 {
-	int			i;
-
 	dump_ind();
 	printf("WHILE ");
 	dump_expr(stmt->cond);
 	printf("\n");
 
-	dump_indent += 2;
-	for (i = 0; i < stmt->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(stmt->body);
 
 	dump_ind();
 	printf("    ENDWHILE\n");
 }
 
 static void
-dump_fori(PLpgSQL_stmt_fori * stmt)
+dump_fori(PLpgSQL_stmt_fori *stmt)
 {
-	int			i;
-
 	dump_ind();
 	printf("FORI %s %s\n", stmt->var->refname, (stmt->reverse) ? "REVERSE" : "NORMAL");
 
@@ -695,59 +798,57 @@ dump_fori(PLpgSQL_stmt_fori * stmt)
 	printf("    upper = ");
 	dump_expr(stmt->upper);
 	printf("\n");
-
-	for (i = 0; i < stmt->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->body->stmts[i]));
+	dump_ind();
+	printf("    step = ");
+	dump_expr(stmt->step);
+	printf("\n");
 	dump_indent -= 2;
 
+	dump_stmts(stmt->body);
+
 	dump_ind();
 	printf("    ENDFORI\n");
 }
 
 static void
-dump_fors(PLpgSQL_stmt_fors * stmt)
+dump_fors(PLpgSQL_stmt_fors *stmt)
 {
-	int			i;
-
 	dump_ind();
 	printf("FORS %s ", (stmt->rec != NULL) ? stmt->rec->refname : stmt->row->refname);
 	dump_expr(stmt->query);
 	printf("\n");
 
-	dump_indent += 2;
-	for (i = 0; i < stmt->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->body->stmts[i]));
-	dump_indent -= 2;
+	dump_stmts(stmt->body);
 
 	dump_ind();
 	printf("    ENDFORS\n");
 }
 
 static void
-dump_select(PLpgSQL_stmt_select * stmt)
+dump_forc(PLpgSQL_stmt_forc *stmt)
 {
 	dump_ind();
-	printf("SELECT ");
-	dump_expr(stmt->query);
-	printf("\n");
+	printf("FORC %s ", stmt->rec->refname);
+	printf("curvar=%d\n", stmt->curvar);
 
 	dump_indent += 2;
-	if (stmt->rec != NULL)
+	if (stmt->argquery != NULL)
 	{
 		dump_ind();
-		printf("    target = %d %s\n", stmt->rec->recno, stmt->rec->refname);
-	}
-	if (stmt->row != NULL)
-	{
-		dump_ind();
-		printf("    target = %d %s\n", stmt->row->rowno, stmt->row->refname);
+		printf("  arguments = ");
+		dump_expr(stmt->argquery);
+		printf("\n");
 	}
 	dump_indent -= 2;
 
+	dump_stmts(stmt->body);
+
+	dump_ind();
+	printf("    ENDFORC\n");
 }
 
 static void
-dump_open(PLpgSQL_stmt_open * stmt)
+dump_open(PLpgSQL_stmt_open *stmt)
 {
 	dump_ind();
 	printf("OPEN curvar=%d\n", stmt->curvar);
@@ -779,35 +880,78 @@ dump_open(PLpgSQL_stmt_open * stmt)
 }
 
 static void
-dump_fetch(PLpgSQL_stmt_fetch * stmt)
+dump_fetch(PLpgSQL_stmt_fetch *stmt)
 {
 	dump_ind();
-	printf("FETCH curvar=%d\n", stmt->curvar);
 
-	dump_indent += 2;
-	if (stmt->rec != NULL)
+	if (!stmt->is_move)
 	{
-		dump_ind();
-		printf("    target = %d %s\n", stmt->rec->recno, stmt->rec->refname);
-	}
-	if (stmt->row != NULL)
-	{
-		dump_ind();
-		printf("    target = %d %s\n", stmt->row->rowno, stmt->row->refname);
-	}
-	dump_indent -= 2;
+		printf("FETCH curvar=%d\n", stmt->curvar);
+		dump_cursor_direction(stmt);
 
+		dump_indent += 2;
+		if (stmt->rec != NULL)
+		{
+			dump_ind();
+			printf("    target = %d %s\n", stmt->rec->dno, stmt->rec->refname);
+		}
+		if (stmt->row != NULL)
+		{
+			dump_ind();
+			printf("    target = %d %s\n", stmt->row->dno, stmt->row->refname);
+		}
+		dump_indent -= 2;
+	}
+	else
+	{
+		printf("MOVE curvar=%d\n", stmt->curvar);
+		dump_cursor_direction(stmt);
+	}
 }
 
 static void
-dump_close(PLpgSQL_stmt_close * stmt)
+dump_cursor_direction(PLpgSQL_stmt_fetch *stmt)
+{
+	dump_indent += 2;
+	dump_ind();
+	switch (stmt->direction)
+	{
+		case FETCH_FORWARD:
+			printf("    FORWARD ");
+			break;
+		case FETCH_BACKWARD:
+			printf("    BACKWARD ");
+			break;
+		case FETCH_ABSOLUTE:
+			printf("    ABSOLUTE ");
+			break;
+		case FETCH_RELATIVE:
+			printf("    RELATIVE ");
+			break;
+		default:
+			printf("??? unknown cursor direction %d", stmt->direction);
+	}
+
+	if (stmt->expr)
+	{
+		dump_expr(stmt->expr);
+		printf("\n");
+	}
+	else
+		printf("%d\n", stmt->how_many);
+
+	dump_indent -= 2;
+}
+
+static void
+dump_close(PLpgSQL_stmt_close *stmt)
 {
 	dump_ind();
 	printf("CLOSE curvar=%d\n", stmt->curvar);
 }
 
 static void
-dump_perform(PLpgSQL_stmt_perform * stmt)
+dump_perform(PLpgSQL_stmt_perform *stmt)
 {
 	dump_ind();
 	printf("PERFORM expr = ");
@@ -816,10 +960,12 @@ dump_perform(PLpgSQL_stmt_perform * stmt)
 }
 
 static void
-dump_exit(PLpgSQL_stmt_exit * stmt)
+dump_exit(PLpgSQL_stmt_exit *stmt)
 {
 	dump_ind();
-	printf("EXIT lbl='%s'", stmt->label);
+	printf("%s", stmt->is_exit ? "EXIT" : "CONTINUE");
+	if (stmt->label != NULL)
+		printf(" label='%s'", stmt->label);
 	if (stmt->cond != NULL)
 	{
 		printf(" WHEN ");
@@ -829,101 +975,243 @@ dump_exit(PLpgSQL_stmt_exit * stmt)
 }
 
 static void
-dump_return(PLpgSQL_stmt_return * stmt)
+dump_return(PLpgSQL_stmt_return *stmt)
 {
 	dump_ind();
 	printf("RETURN ");
-	if (stmt->retrecno >= 0)
-		printf("record %d", stmt->retrecno);
-	else if (stmt->retrowno >= 0)
-		printf("row %d", stmt->retrowno);
-	else if (stmt->expr == NULL)
-		printf("NULL");
-	else
+	if (stmt->retvarno >= 0)
+		printf("variable %d", stmt->retvarno);
+	else if (stmt->expr != NULL)
 		dump_expr(stmt->expr);
+	else
+		printf("NULL");
 	printf("\n");
 }
 
 static void
-dump_return_next(PLpgSQL_stmt_return_next * stmt)
+dump_return_next(PLpgSQL_stmt_return_next *stmt)
 {
 	dump_ind();
 	printf("RETURN NEXT ");
-	if (stmt->rec != NULL)
-		printf("target = %d %s\n", stmt->rec->recno, stmt->rec->refname);
-	else if (stmt->row != NULL)
-		printf("target = %d %s\n", stmt->row->rowno, stmt->row->refname);
+	if (stmt->retvarno >= 0)
+		printf("variable %d", stmt->retvarno);
 	else if (stmt->expr != NULL)
 		dump_expr(stmt->expr);
+	else
+		printf("NULL");
 	printf("\n");
 }
 
 static void
-dump_raise(PLpgSQL_stmt_raise * stmt)
+dump_return_query(PLpgSQL_stmt_return_query *stmt)
 {
-	int			i;
+	dump_ind();
+	if (stmt->query)
+	{
+		printf("RETURN QUERY ");
+		dump_expr(stmt->query);
+		printf("\n");
+	}
+	else
+	{
+		printf("RETURN QUERY EXECUTE ");
+		dump_expr(stmt->dynquery);
+		printf("\n");
+		if (stmt->params != NIL)
+		{
+			ListCell   *lc;
+			int			i;
+
+			dump_indent += 2;
+			dump_ind();
+			printf("    USING\n");
+			dump_indent += 2;
+			i = 1;
+			foreach(lc, stmt->params)
+			{
+				dump_ind();
+				printf("    parameter $%d: ", i++);
+				dump_expr((PLpgSQL_expr *) lfirst(lc));
+				printf("\n");
+			}
+			dump_indent -= 4;
+		}
+	}
+}
+
+static void
+dump_raise(PLpgSQL_stmt_raise *stmt)
+{
+	ListCell   *lc;
+	int			i = 0;
 
 	dump_ind();
-	printf("RAISE '%s'", stmt->message);
-	for (i = 0; i < stmt->nparams; i++)
-		printf(" %d", stmt->params[i]);
+	printf("RAISE level=%d", stmt->elog_level);
+	if (stmt->condname)
+		printf(" condname='%s'", stmt->condname);
+	if (stmt->message)
+		printf(" message='%s'", stmt->message);
 	printf("\n");
+	dump_indent += 2;
+	foreach(lc, stmt->params)
+	{
+		dump_ind();
+		printf("    parameter %d: ", i++);
+		dump_expr((PLpgSQL_expr *) lfirst(lc));
+		printf("\n");
+	}
+	if (stmt->options)
+	{
+		dump_ind();
+		printf("    USING\n");
+		dump_indent += 2;
+		foreach(lc, stmt->options)
+		{
+			PLpgSQL_raise_option *opt = (PLpgSQL_raise_option *) lfirst(lc);
+
+			dump_ind();
+			switch (opt->opt_type)
+			{
+				case PLPGSQL_RAISEOPTION_ERRCODE:
+					printf("    ERRCODE = ");
+					break;
+				case PLPGSQL_RAISEOPTION_MESSAGE:
+					printf("    MESSAGE = ");
+					break;
+				case PLPGSQL_RAISEOPTION_DETAIL:
+					printf("    DETAIL = ");
+					break;
+				case PLPGSQL_RAISEOPTION_HINT:
+					printf("    HINT = ");
+					break;
+			}
+			dump_expr(opt->expr);
+			printf("\n");
+		}
+		dump_indent -= 2;
+	}
+	dump_indent -= 2;
 }
 
 static void
-dump_execsql(PLpgSQL_stmt_execsql * stmt)
+dump_execsql(PLpgSQL_stmt_execsql *stmt)
 {
 	dump_ind();
 	printf("EXECSQL ");
 	dump_expr(stmt->sqlstmt);
 	printf("\n");
+
+	dump_indent += 2;
+	if (stmt->rec != NULL)
+	{
+		dump_ind();
+		printf("    INTO%s target = %d %s\n",
+			   stmt->strict ? " STRICT" : "",
+			   stmt->rec->dno, stmt->rec->refname);
+	}
+	if (stmt->row != NULL)
+	{
+		dump_ind();
+		printf("    INTO%s target = %d %s\n",
+			   stmt->strict ? " STRICT" : "",
+			   stmt->row->dno, stmt->row->refname);
+	}
+	dump_indent -= 2;
 }
 
 static void
-dump_dynexecute(PLpgSQL_stmt_dynexecute * stmt)
+dump_dynexecute(PLpgSQL_stmt_dynexecute *stmt)
 {
 	dump_ind();
 	printf("EXECUTE ");
 	dump_expr(stmt->query);
 	printf("\n");
+
+	dump_indent += 2;
+	if (stmt->rec != NULL)
+	{
+		dump_ind();
+		printf("    INTO%s target = %d %s\n",
+			   stmt->strict ? " STRICT" : "",
+			   stmt->rec->dno, stmt->rec->refname);
+	}
+	if (stmt->row != NULL)
+	{
+		dump_ind();
+		printf("    INTO%s target = %d %s\n",
+			   stmt->strict ? " STRICT" : "",
+			   stmt->row->dno, stmt->row->refname);
+	}
+	if (stmt->params != NIL)
+	{
+		ListCell   *lc;
+		int			i;
+
+		dump_ind();
+		printf("    USING\n");
+		dump_indent += 2;
+		i = 1;
+		foreach(lc, stmt->params)
+		{
+			dump_ind();
+			printf("    parameter %d: ", i++);
+			dump_expr((PLpgSQL_expr *) lfirst(lc));
+			printf("\n");
+		}
+		dump_indent -= 2;
+	}
+	dump_indent -= 2;
 }
 
 static void
-dump_dynfors(PLpgSQL_stmt_dynfors * stmt)
+dump_dynfors(PLpgSQL_stmt_dynfors *stmt)
 {
-	int			i;
-
 	dump_ind();
-	printf("FORS %s EXECUTE ", (stmt->rec != NULL) ? stmt->rec->refname : stmt->row->refname);
+	printf("FORS %s EXECUTE ",
+		   (stmt->rec != NULL) ? stmt->rec->refname : stmt->row->refname);
 	dump_expr(stmt->query);
 	printf("\n");
+	if (stmt->params != NIL)
+	{
+		ListCell   *lc;
+		int			i;
 
-	dump_indent += 2;
-	for (i = 0; i < stmt->body->stmts_used; i++)
-		dump_stmt((PLpgSQL_stmt *) (stmt->body->stmts[i]));
-	dump_indent -= 2;
-
+		dump_indent += 2;
+		dump_ind();
+		printf("    USING\n");
+		dump_indent += 2;
+		i = 1;
+		foreach(lc, stmt->params)
+		{
+			dump_ind();
+			printf("    parameter $%d: ", i++);
+			dump_expr((PLpgSQL_expr *) lfirst(lc));
+			printf("\n");
+		}
+		dump_indent -= 4;
+	}
+	dump_stmts(stmt->body);
 	dump_ind();
 	printf("    ENDFORS\n");
 }
 
 static void
-dump_getdiag(PLpgSQL_stmt_getdiag * stmt)
+dump_getdiag(PLpgSQL_stmt_getdiag *stmt)
 {
-	int			i;
+	ListCell   *lc;
 
 	dump_ind();
 	printf("GET DIAGNOSTICS ");
-	for (i = 0; i < stmt->ndtitems; i++)
+	foreach(lc, stmt->diag_items)
 	{
-		PLpgSQL_diag_item *dtitem = &stmt->dtitems[i];
+		PLpgSQL_diag_item *diag_item = (PLpgSQL_diag_item *) lfirst(lc);
 
-		if (i != 0)
+		if (lc != list_head(stmt->diag_items))
 			printf(", ");
 
-		printf("{var %d} = ", dtitem->target);
+		printf("{var %d} = ", diag_item->target);
 
-		switch (dtitem->item)
+		switch (diag_item->kind)
 		{
 			case PLPGSQL_GETDIAG_ROW_COUNT:
 				printf("ROW_COUNT");
@@ -942,7 +1230,7 @@ dump_getdiag(PLpgSQL_stmt_getdiag * stmt)
 }
 
 static void
-dump_expr(PLpgSQL_expr * expr)
+dump_expr(PLpgSQL_expr *expr)
 {
 	int			i;
 
@@ -962,7 +1250,7 @@ dump_expr(PLpgSQL_expr * expr)
 }
 
 void
-plpgsql_dumptree(PLpgSQL_function * func)
+plpgsql_dumptree(PLpgSQL_function *func)
 {
 	int			i;
 	PLpgSQL_datum *d;
@@ -970,7 +1258,7 @@ plpgsql_dumptree(PLpgSQL_function * func)
 	printf("\nExecution tree of successfully compiled PL/pgSQL function %s:\n",
 		   func->fn_name);
 
-	printf("\nFunctions data area:\n");
+	printf("\nFunction's data area:\n");
 	for (i = 0; i < func->ndatums; i++)
 	{
 		d = func->datums[i];
@@ -1045,10 +1333,11 @@ plpgsql_dumptree(PLpgSQL_function * func)
 				printf("??? unknown data type %d\n", d->dtype);
 		}
 	}
-	printf("\nFunctions statements:\n");
+	printf("\nFunction's statements:\n");
 
 	dump_indent = 0;
 	printf("%3d:", func->action->lineno);
 	dump_block(func->action);
 	printf("\nEnd of execution tree of function %s\n\n", func->fn_name);
+	fflush(stdout);
 }
